import os
import uuid
from typing import Any, Dict, Generator, List

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Concept, ConceptRelation
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.base import BaseClient
from clarifai.client.dataset import Dataset
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.client.model import Model
from clarifai.client.module import Module
from clarifai.client.pipeline import Pipeline
from clarifai.client.pipeline_step import PipelineStep
from clarifai.client.search import Search
from clarifai.client.workflow import Workflow
from clarifai.constants.model import TRAINABLE_MODEL_TYPES
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import DEFAULT_BASE
from clarifai.utils.logging import display_concept_relations_tree, display_workflow_tree, logger
from clarifai.utils.misc import concept_relations_accumulation
from clarifai.workflows.utils import get_yaml_output_info_proto, is_same_yaml_model
from clarifai.workflows.validate import validate


class App(Lister, BaseClient):
    """App is a class that provides access to Clarifai API endpoints related to App information."""

    def __init__(
        self,
        url: str = None,
        app_id: str = None,
        user_id: str = None,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
        **kwargs,
    ):
        """Initializes an App object.

        Args:
            url (str): The URL to initialize the app object.
            app_id (str): The App ID for the App to interact with.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the App.
                - name (str): The name of the app.
                - description (str): The description of the app.
        """
        if url and app_id:
            raise UserError("You can only specify one of url or app_id.")
        if url:
            user_id, app_id = ClarifaiUrlHelper.split_clarifai_app_url(url)
            kwargs = {'user_id': user_id}
        if user_id:
            kwargs = {'user_id': user_id}

        self.kwargs = {**kwargs, 'id': app_id}
        self.app_info = resources_pb2.App(**self.kwargs)
        self.logger = logger
        BaseClient.__init__(
            self,
            user_id=self.user_id,
            app_id=self.id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

    def list_datasets(
        self, page_no: int = None, per_page: int = None
    ) -> Generator[Dataset, None, None]:
        """Lists all the datasets for the app.

        Args:
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Dataset: Dataset objects for the datasets in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_datasets = list(app.list_datasets())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(
            user_app_id=self.user_app_id,
        )
        all_datasets_info = self.list_pages_generator(
            self.STUB.ListDatasets,
            service_pb2.ListDatasetsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )
        for dataset_info in all_datasets_info:
            if 'version' in dataset_info:
                dataset_info['version'].pop('metrics', None)
                dataset_info['version'].pop('export_info', None)
            yield Dataset.from_auth_helper(auth=self.auth_helper, **dataset_info)

    def list_models(
        self,
        filter_by: Dict[str, Any] = {},
        only_in_app: bool = True,
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[Model, None, None]:
        """Lists all the available models for the user.

        Args:
            filter_by (dict): A dictionary of filters to apply to the list of models.
            only_in_app (bool): If True, only return models that are in the app.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Model: Model objects for the models in the app.

        Example:
            >>> from clarifai.client.user import User
            >>> app = User(user_id="user_id").app(app_id="app_id")
            >>> all_models = list(app.list_models())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        all_models_info = self.list_pages_generator(
            self.STUB.ListModels,
            service_pb2.ListModelsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for model_info in all_models_info:
            if 'model_version' not in list(model_info.keys()):
                continue
            if only_in_app:
                if model_info['app_id'] != self.id:
                    continue
            yield Model.from_auth_helper(auth=self.auth_helper, **model_info)

    def list_workflows(
        self,
        filter_by: Dict[str, Any] = {},
        only_in_app: bool = True,
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[Workflow, None, None]:
        """Lists all the available workflows for the user.

        Args:
            filter_by (dict): A dictionary of filters to apply to the list of workflows.
            only_in_app (bool): If True, only return workflows that are in the app.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Workflow: Workflow objects for the workflows in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_workflows = list(app.list_workflows())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        all_workflows_info = self.list_pages_generator(
            self.STUB.ListWorkflows,
            service_pb2.ListWorkflowsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for workflow_info in all_workflows_info:
            if only_in_app:
                if workflow_info['app_id'] != self.id:
                    continue
            yield Workflow.from_auth_helper(auth=self.auth_helper, **workflow_info)

    def list_pipelines(
        self,
        filter_by: Dict[str, Any] = {},
        only_in_app: bool = True,
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[dict, None, None]:
        """Lists all the pipelines for the user.

        Args:
            filter_by (dict): A dictionary of filters to apply to the list of pipelines.
            only_in_app (bool): If True, only return pipelines that are in the app.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Dict: Dictionaries containing information about the pipelines.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_pipelines = list(app.list_pipelines())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        all_pipelines_info = self.list_pages_generator(
            self.STUB.ListPipelines,
            service_pb2.ListPipelinesRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for pipeline_info in all_pipelines_info:
            pipeline = self._process_pipeline_info(
                pipeline_info, self.auth_helper, self.id, only_in_app
            )
            if pipeline is not None:
                yield pipeline

    @staticmethod
    def _process_pipeline_info(pipeline_info, auth_helper, app_id=None, only_in_app=False):
        """Helper method to process pipeline info and create Pipeline objects.

        Args:
            pipeline_info: Raw pipeline info from API
            auth_helper: Auth helper instance
            app_id: App ID to filter by (if only_in_app is True)
            only_in_app: Whether to filter by app_id

        Returns:
            Pipeline object or None if filtered out
        """
        if only_in_app and app_id:
            if pipeline_info.get('app_id') != app_id:
                return None

        # Map API field names to constructor parameter names
        pipeline_kwargs = pipeline_info.copy()
        if 'id' in pipeline_kwargs:
            pipeline_kwargs['pipeline_id'] = pipeline_kwargs.pop('id')
        if 'pipeline_version' in pipeline_kwargs:
            pipeline_version = pipeline_kwargs.pop('pipeline_version')
            pipeline_kwargs['pipeline_version_id'] = pipeline_version.get('id', '')

        return Pipeline.from_auth_helper(auth=auth_helper, **pipeline_kwargs)

    @staticmethod
    def _process_pipeline_step_info(
        pipeline_step_info, auth_helper, app_id=None, only_in_app=False
    ):
        """Helper method to process pipeline step info and create PipelineStep objects.

        Args:
            pipeline_step_info: Raw pipeline step info from API
            auth_helper: Auth helper instance
            app_id: App ID to filter by (if only_in_app is True)
            only_in_app: Whether to filter by app_id

        Returns:
            PipelineStep object or None if filtered out
        """
        if only_in_app and app_id:
            if pipeline_step_info.get('app_id') != app_id:
                return None

        # Map API field names to constructor parameter names
        step_kwargs = pipeline_step_info.copy()
        if 'pipeline_step' in step_kwargs:
            pipeline_step = step_kwargs.pop('pipeline_step')
            step_kwargs['pipeline_step_id'] = pipeline_step.get('id', '')

        return PipelineStep.from_auth_helper(auth=auth_helper, **step_kwargs)

    def list_pipeline_steps(
        self,
        pipeline_id: str = None,
        filter_by: Dict[str, Any] = {},
        only_in_app: bool = True,
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[dict, None, None]:
        """Lists all the pipeline steps for the user.

        Args:
            pipeline_id (str): If provided, only list pipeline steps from this pipeline.
            filter_by (dict): A dictionary of filters to apply to the list of pipeline steps.
            only_in_app (bool): If True, only return pipeline steps that are in the app.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Dict: Dictionaries containing information about the pipeline steps.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_pipeline_steps = list(app.list_pipeline_steps())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        if pipeline_id:
            request_data['pipeline_id'] = pipeline_id

        all_pipeline_steps_info = self.list_pages_generator(
            self.STUB.ListPipelineStepVersions,
            service_pb2.ListPipelineStepVersionsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for pipeline_step_info in all_pipeline_steps_info:
            pipeline_step = self._process_pipeline_step_info(
                pipeline_step_info, self.auth_helper, self.id, only_in_app
            )
            if pipeline_step is not None:
                yield pipeline_step

    def list_modules(
        self,
        filter_by: Dict[str, Any] = {},
        only_in_app: bool = True,
        page_no: int = None,
        per_page: int = None,
    ) -> Generator[Module, None, None]:
        """Lists all the available modules for the user.

        Args:
            filter_by (dict): A dictionary of filters to apply to the list of modules.
            only_in_app (bool): If True, only return modules that are in the app.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Module: Module objects for the modules in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_modules = list(app.list_modules())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        all_modules_info = self.list_pages_generator(
            self.STUB.ListModules,
            service_pb2.ListModulesRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for module_info in all_modules_info:
            if only_in_app:
                if module_info['app_id'] != self.id:
                    continue
            yield Module.from_auth_helper(auth=self.auth_helper, **module_info)

    def list_installed_module_versions(
        self, filter_by: Dict[str, Any] = {}, page_no: int = None, per_page: int = None
    ) -> Generator[Module, None, None]:
        """Lists all installed module versions in the app.

        Args:
            filter_by (dict): A dictionary of filters to apply to the list of installed module versions.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Module: Module objects for the installed module versions in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_installed_module_versions = list(app.list_installed_module_versions())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, **filter_by)
        all_imv_infos = self.list_pages_generator(
            self.STUB.ListInstalledModuleVersions,
            service_pb2.ListInstalledModuleVersionsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )
        for imv_info in all_imv_infos:
            del imv_info['deploy_url']
            del imv_info['installed_module_version_id']  # TODO: remove this after the backend fix
            yield Module.from_auth_helper(
                auth=self.auth_helper,
                module_id=imv_info['module_version']['module_id'],
                **imv_info,
            )

    def get_input_count(self) -> int:
        """Get count of all the inputs in the app.

        Returns:
            input_count: count of inputs in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> input_count = app.get_input_count()
        """
        request = service_pb2.GetInputCountRequest(user_app_id=self.user_app_id)
        response = self._grpc_request(self.STUB.GetInputCount, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nGetting App input Counts\n%s", response.status)

        return response.counts.processed

    def list_concepts(
        self, page_no: int = None, per_page: int = None
    ) -> Generator[Concept, None, None]:
        """Lists all the concepts for the app.
        Args:
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Concept: Concepts in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_concepts = list(app.list_concepts())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id)
        all_concepts_infos = self.list_pages_generator(
            self.STUB.ListConcepts,
            service_pb2.ListConceptsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )
        for concept_info in all_concepts_infos:
            concept_info['id'] = concept_info.pop('concept_id')
            yield Concept(**concept_info)

    def search_concept_relations(
        self,
        concept_id: str = None,
        predicate: str = None,
        page_no: int = None,
        per_page: int = None,
        show_tree: bool = False,
    ) -> Generator[ConceptRelation, None, None]:
        """List all the concept relations of the app.

        Args:
            concept_id (str, optional): The concept ID to filter the concept relations.
            predicate (str, optional): Type of relation to filter the concept relations. For ex : 'hypernym', 'hyponym' or 'synonym'
            page_no (int, optional): The page number to list.
            per_page (int, optional): The number of items per page.
            show_tree (bool, optional): If True, prints out rich tree representation of concept relations.

        Yields:
            ConceptRelation: ConceptRelations in the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> all_concept_relations = list(app.search_concept_relations())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(
            user_app_id=self.user_app_id, concept_id=concept_id, predicate=predicate
        )
        all_concept_relations_infos = self.list_pages_generator(
            self.STUB.ListConceptRelations,
            service_pb2.ListConceptRelationsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )
        relations_dict = {}
        for concept_relation_info in all_concept_relations_infos:
            if show_tree:
                current_subject_concept = concept_relation_info['subject_concept']['id']
                current_object_concept = concept_relation_info['object_concept']['id']
                current_predicate = concept_relation_info['predicate']
                relations_dict = concept_relations_accumulation(
                    relations_dict,
                    current_subject_concept,
                    current_object_concept,
                    current_predicate,
                )
            concept_relation_info['id'] = concept_relation_info.pop('concept_relation_id')
            yield ConceptRelation(**concept_relation_info)
        if show_tree:
            display_concept_relations_tree(relations_dict)

    def list_trainable_model_types(self) -> List[str]:
        """Lists all the trainable model types.

        Returns:
            templates (List): List all the trainable model types.

        Example:
            >>> from clarifai.client.app import App
            >>> print(app.list_trainable_model_types())
        """
        return TRAINABLE_MODEL_TYPES

    def create_dataset(self, dataset_id: str, **kwargs) -> Dataset:
        """Creates a dataset for the app.

        Args:
            dataset_id (str): The dataset ID for the dataset to create.
            **kwargs: Additional keyword arguments to be passed to the Dataset.

        Returns:
            Dataset: A Dataset object for the specified dataset ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> dataset = app.create_dataset(dataset_id="dataset_id")
        """
        request = service_pb2.PostDatasetsRequest(
            user_app_id=self.user_app_id, datasets=[resources_pb2.Dataset(id=dataset_id, **kwargs)]
        )
        response = self._grpc_request(self.STUB.PostDatasets, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset created\n%s", response.status)

        return Dataset.from_auth_helper(self.auth_helper, dataset_id=dataset_id, **kwargs)

    def create_model(self, model_id: str, **kwargs) -> Model:
        """Creates a model for the app.

        Args:
            model_id (str): The model ID for the model to create.
            **kwargs: Additional keyword arguments to be passed to the Model.

        Returns:
            Model: A Model object for the specified model ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> model = app.create_model(model_id="model_id")
        """
        request = service_pb2.PostModelsRequest(
            user_app_id=self.user_app_id, models=[resources_pb2.Model(id=model_id, **kwargs)]
        )
        response = self._grpc_request(self.STUB.PostModels, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info(f"Model with ID '{model_id}' is created:\n{response.status}")
        kwargs.update(
            {
                'model_id': model_id,
                'model_type_id': response.model.model_type_id,
            }
        )

        return Model.from_auth_helper(auth=self.auth_helper, **kwargs)

    def _process_workflow_config(self, config_filepath: str):
        with open(config_filepath, 'r') as file:
            workflow_config = yaml.safe_load(file)

        workflow_config = validate(workflow_config)
        workflow = workflow_config['workflow']

        # Get all model objects from the workflow nodes.
        all_models = []
        for node in workflow['nodes']:
            output_info = get_yaml_output_info_proto(node['model'].get('output_info', None))
            try:
                model = self.model(
                    model_id=node['model']['model_id'],
                    model_version={"id": node['model'].get('model_version_id', "")},
                    user_id=node['model'].get('user_id', self.user_app_id.user_id),
                    app_id=node['model'].get('app_id', self.user_app_id.app_id),
                )
            except Exception as e:
                if "Model does not exist" in str(e):
                    model = self.create_model(
                        **{k: v for k, v in node['model'].items() if k != 'output_info'}
                    )
                    model_version = model.create_version(output_info=output_info)
                    all_models.append(model_version.model_info)
                    continue

            # If the model version ID is specified, or if the yaml model is the same as the one in the api
            if node["model"].get("model_version_id", "") or is_same_yaml_model(
                model.model_info, node["model"]
            ):
                all_models.append(model.model_info)
            else:  # Create a new model version
                model = model.create_version(output_info=output_info)
                all_models.append(model.model_info)

        # Convert nodes to resources_pb2.WorkflowNodes.
        nodes = []
        for i, yml_node in enumerate(workflow['nodes']):
            node = resources_pb2.WorkflowNode(
                id=yml_node['id'],
                model=all_models[i],
            )
            # Add node inputs if they exist, i.e. if these nodes do not connect directly to the input.
            if yml_node.get("node_inputs"):
                for ni in yml_node.get("node_inputs"):
                    node.node_inputs.append(resources_pb2.NodeInput(node_id=ni['node_id']))
            nodes.append(node)

        return workflow, nodes

    def create_workflow(
        self, config_filepath: str, generate_new_id: bool = False, display: bool = True
    ) -> Workflow:
        """Creates a workflow for the app.

        Args:
            config_filepath (str): The path to the yaml workflow config file.
            generate_new_id (bool): If True, generate a new workflow ID.
            display (bool): If True, display the workflow nodes tree.

        Returns:
            Workflow: A Workflow object for the specified workflow config.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> workflow = app.create_workflow(config_filepath="config.yml")
        """
        if not os.path.exists(config_filepath):
            raise UserError(f"Workflow config file not found at {config_filepath}")

        workflow, nodes = self._process_workflow_config(config_filepath)

        workflow_id = workflow['id']
        if generate_new_id:
            workflow_id = str(uuid.uuid4())

        # Create the workflow.
        request = service_pb2.PostWorkflowsRequest(
            user_app_id=self.user_app_id,
            workflows=[resources_pb2.Workflow(id=workflow_id, nodes=nodes)],
        )

        response = self._grpc_request(self.STUB.PostWorkflows, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nWorkflow created\n%s", response.status)

        dict_response = MessageToDict(response, preserving_proto_field_name=True)
        # Display the workflow nodes tree.
        wf = dict_response["workflows"][0]
        if display:
            display_workflow_tree(wf["nodes"])
        kwargs = self.process_response_keys(wf, "workflow")

        return Workflow.from_auth_helper(auth=self.auth_helper, **kwargs)

    def create_module(self, module_id: str, description: str, **kwargs) -> Module:
        """Creates a module for the app.

        Args:
            module_id (str): The module ID for the module to create.
            description (str): The description of the module to create.
            **kwargs: Additional keyword arguments to be passed to the module.

        Returns:
            Module: A Module object for the specified module ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> module = app.create_module(module_id="module_id")
        """
        request = service_pb2.PostModulesRequest(
            user_app_id=self.user_app_id,
            modules=[resources_pb2.Module(id=module_id, description=description, **kwargs)],
        )
        response = self._grpc_request(self.STUB.PostModules, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nModule created\n%s", response.status)

        return Module.from_auth_helper(auth=self.auth_helper, module_id=module_id, **kwargs)

    def create_concepts(self, concept_ids: List[str], concepts: List[str] = []) -> None:
        """Add concepts to the app.

        Args:
            concept_ids (List[str]): List of concept IDs of concepts to add to the app.
            concepts (List[str], optional): List of concepts names of concepts to add to the app.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.add_concepts(concept_ids=["concept_id_1", "concept_id_2", ..])
        """
        if not concepts:
            concepts = list(concept_ids)
        concepts_to_add = [
            resources_pb2.Concept(id=concept_id, name=concept)
            for concept_id, concept in zip(concept_ids, concepts)
        ]
        request = service_pb2.PostConceptsRequest(
            user_app_id=self.user_app_id, concepts=concepts_to_add
        )
        response = self._grpc_request(self.STUB.PostConcepts, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nConcepts added\n%s", response.status)

    def create_concept_relations(
        self, subject_concept_id: str, object_concept_ids: List[str], predicates: List[str]
    ) -> None:
        """Creates concept relations between concepts of the app.

        Args:
            subject_concept_id (str): The concept ID of the subject concept.
            object_concept_ids (List[str]): List of concepts IDs of object concepts.
            predicates (List[str]): List of predicates corresponding to each relation between subject and objects.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.create_concept_relation(subject_concept_id="subject_concept_id", object_concept_ids=["object_concept_id_1", "object_concept_id_2", ..], predicates=["predicate_1", "predicate_2", ..])
        """
        assert len(object_concept_ids) == len(predicates)
        subject_relations = [
            resources_pb2.ConceptRelation(
                object_concept=resources_pb2.Concept(id=object_concept_id), predicate=predicate
            )
            for object_concept_id, predicate in zip(object_concept_ids, predicates)
        ]
        request = service_pb2.PostConceptRelationsRequest(
            user_app_id=self.user_app_id,
            concept_id=subject_concept_id,
            concept_relations=subject_relations,
        )
        response = self._grpc_request(self.STUB.PostConceptRelations, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nConcept Relations created\n%s", response.status)

    def dataset(self, dataset_id: str, dataset_version_id: str = None, **kwargs) -> Dataset:
        """Returns a Dataset object for the existing dataset ID.

        Args:
            dataset_id (str): The dataset ID for the dataset to interact with.
            dataset_version_id (str): The version ID for the dataset version to interact with.

        Returns:
            Dataset: A Dataset object for the existing dataset ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> dataset = app.dataset(dataset_id="dataset_id")
        """
        request = service_pb2.GetDatasetRequest(
            user_app_id=self.user_app_id, dataset_id=dataset_id
        )
        response = self._grpc_request(self.STUB.GetDataset, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        dict_response = MessageToDict(response, preserving_proto_field_name=True)
        kwargs = self.process_response_keys(dict_response['dataset'], 'dataset')
        kwargs['version'] = response.dataset.version if response.dataset.version else None
        kwargs['dataset_version_id'] = dataset_version_id
        return Dataset.from_auth_helper(auth=self.auth_helper, **kwargs)

    def model(self, model_id: str, model_version: Dict = {'id': ""}, **kwargs) -> Model:
        """Returns a Model object for the existing model ID.

        Args:
            model_id (str): The model ID for the model to interact with.
            model_version (Dict): The model version ID for the model version to interact with.

        Returns:
            Model: A Model object for the existing model ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> model_v1 = app.model(model_id="model_id", model_version={"id":"model_version_id")
        """
        # Change user_app_id based on whether user_id or app_id is specified.
        if kwargs.get("user_id") or kwargs.get("app_id"):
            request = service_pb2.GetModelRequest(
                user_app_id=self.auth_helper.get_user_app_id_proto(
                    kwargs.get("user_id"), kwargs.get("app_id")
                ),
                model_id=model_id,
                version_id=model_version["id"],
            )
        else:
            request = service_pb2.GetModelRequest(
                user_app_id=self.user_app_id, model_id=model_id, version_id=model_version["id"]
            )
        response = self._grpc_request(self.STUB.GetModel, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        dict_response = MessageToDict(response, preserving_proto_field_name=True)

        kwargs = self.process_response_keys(dict_response['model'], 'model')
        kwargs['model_version'] = (
            response.model.model_version if response.model.model_version else None
        )

        m = Model.from_auth_helper(self.auth_helper, **kwargs)
        return m

    def workflow(self, workflow_id: str, **kwargs) -> Workflow:
        """Returns a workflow object for the existing workflow ID.

        Args:
            workflow_id (str): The workflow ID for the workflow to interact with.

        Returns:
            Workflow: A Workflow object for the existing workflow ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> workflow = app.workflow(workflow_id="workflow_id")
        """
        request = service_pb2.GetWorkflowRequest(
            user_app_id=self.user_app_id, workflow_id=workflow_id
        )
        response = self._grpc_request(self.STUB.GetWorkflow, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        dict_response = MessageToDict(response, preserving_proto_field_name=True)
        kwargs = self.process_response_keys(dict_response['workflow'], "workflow")

        return Workflow.from_auth_helper(auth=self.auth_helper, **kwargs)

    def module(self, module_id: str, **kwargs) -> Module:
        """Returns a Module object for the existing module ID.

        Args:
            module_id (str): The module ID for the module to interact with.
            module_version_id (str): The module version ID for the module version to interact with.

        Returns:
            Module: A Module object for the existing module ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> module = app.module(module_id="module_id")
        """
        request = service_pb2.GetModuleRequest(user_app_id=self.user_app_id, module_id=module_id)
        response = self._grpc_request(self.STUB.GetModule, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        dict_response = MessageToDict(response, preserving_proto_field_name=True)
        kwargs = self.process_response_keys(dict_response['module'], 'module')

        return Module.from_auth_helper(auth=self.auth_helper, **kwargs)

    def inputs(
        self,
    ):
        """Returns an Input object.

        Returns:
            Inputs: An input object.
        """
        return Inputs.from_auth_helper(self.auth_helper)

    def patch_dataset(self, dataset_id: str, action: str = 'merge', **kwargs) -> Dataset:
        """Patches a dataset for the app.

        Args:
            dataset_id (str): The dataset ID for the dataset to create.
            action (str): The action to perform on the dataset (merge/overwrite/remove).
            **kwargs: Additional keyword arguments to be passed to patch the Dataset.

        Returns:
            Dataset: A Dataset object for the specified dataset ID.
        """
        if "visibility" in kwargs:
            kwargs["visibility"] = resources_pb2.Visibility(gettable=kwargs["visibility"])
        if "image_url" in kwargs:
            kwargs["image"] = resources_pb2.Image(url=kwargs.pop("image_url"))
        request = service_pb2.PatchDatasetsRequest(
            user_app_id=self.user_app_id,
            datasets=[resources_pb2.Dataset(id=dataset_id, **kwargs)],
            action=action,
        )
        response = self._grpc_request(self.STUB.PatchDatasets, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset patched\n%s", response.status)

        return Dataset.from_auth_helper(self.auth_helper, dataset_id=dataset_id, **kwargs)

    def patch_model(self, model_id: str, action: str = 'merge', **kwargs) -> Model:
        """Patches a model of the app.

        Args:
            model_id (str): The model ID of the model to patch.
            action (str): The action to perform on the model (merge/overwrite/remove).
            **kwargs: Additional keyword arguments to be passed to patch the Model.

        Returns:
            Model: A Model object of the specified model ID.
        """
        if "visibility" in kwargs:
            kwargs["visibility"] = resources_pb2.Visibility(gettable=kwargs["visibility"])
        if "image_url" in kwargs:
            kwargs["image"] = resources_pb2.Image(url=kwargs.pop("image_url"))
        request = service_pb2.PatchModelsRequest(
            user_app_id=self.user_app_id,
            models=[resources_pb2.Model(id=model_id, **kwargs)],
            action=action,
        )
        response = self._grpc_request(self.STUB.PatchModels, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nModel %s patched successfully\n%s", model_id, response.status)
        kwargs.update(
            {
                'model_id': model_id,
            }
        )

        return Model.from_auth_helper(self.auth_helper, **kwargs)

    def patch_workflow(
        self, workflow_id: str, action: str = 'merge', config_filepath: str = None, **kwargs
    ) -> Workflow:
        """Patches a workflow of the app.

        Args:
            workflow_id (str): The Workflow ID of the workflow to patch.
            action (str): The action to perform on the workflow (merge/overwrite/remove).
            config_filepath (str): The path to the yaml workflow config file.
            **kwargs: Additional keyword arguments to be passed to patch the Workflow.

        Returns:
            Workflow: A Workflow object of the specified workflow ID.
        """
        if config_filepath:
            if not os.path.exists(config_filepath):
                raise UserError(f"Workflow config file not found at {config_filepath}")
            _, kwargs['nodes'] = self._process_workflow_config(config_filepath)
        if "visibility" in kwargs:
            kwargs["visibility"] = resources_pb2.Visibility(gettable=kwargs["visibility"])
        if "image_url" in kwargs:
            kwargs["image"] = resources_pb2.Image(url=kwargs.pop("image_url"))

        request = service_pb2.PatchWorkflowsRequest(
            user_app_id=self.user_app_id,
            workflows=[resources_pb2.Workflow(id=workflow_id, **kwargs)],
            action=action,
        )
        response = self._grpc_request(self.STUB.PatchWorkflows, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nWorkflow patched\n%s", response.status)
        kwargs.update(
            {
                'workflow_id': workflow_id,
            }
        )

        return Workflow.from_auth_helper(self.auth_helper, **kwargs)

    def delete_dataset(self, dataset_id: str) -> None:
        """Deletes an dataset for the user.

        Args:
            dataset_id (str): The dataset ID for the app to delete.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.delete_dataset(dataset_id="dataset_id")
        """
        request = service_pb2.DeleteDatasetsRequest(
            user_app_id=self.user_app_id, dataset_ids=[dataset_id]
        )
        response = self._grpc_request(self.STUB.DeleteDatasets, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset Deleted\n%s", response.status)

    def delete_model(self, model_id: str) -> None:
        """Deletes an model for the user.

        Args:
            model_id (str): The model ID for the app to delete.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.delete_model(model_id="model_id")
        """
        request = service_pb2.DeleteModelsRequest(user_app_id=self.user_app_id, ids=[model_id])
        response = self._grpc_request(self.STUB.DeleteModels, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nModel Deleted\n%s", response.status)

    def delete_workflow(self, workflow_id: str) -> None:
        """Deletes an workflow for the user.

        Args:
            workflow_id (str): The workflow ID for the app to delete.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.delete_workflow(workflow_id="workflow_id")
        """
        request = service_pb2.DeleteWorkflowsRequest(
            user_app_id=self.user_app_id, ids=[workflow_id]
        )
        response = self._grpc_request(self.STUB.DeleteWorkflows, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nWorkflow Deleted\n%s", response.status)

    def delete_module(self, module_id: str) -> None:
        """Deletes an module for the user.

        Args:
            module_id (str): The module ID for the app to delete.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.delete_module(module_id="module_id")
        """
        request = service_pb2.DeleteModulesRequest(user_app_id=self.user_app_id, ids=[module_id])
        response = self._grpc_request(self.STUB.DeleteModules, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nModule Deleted\n%s", response.status)

    def delete_concept_relations(
        self, concept_id: str, concept_relation_ids: List[str] = []
    ) -> None:
        """Delete concept relations of a concept of the app.

        Args:
            concept_id (str): The concept ID of the concept to delete relations for.
            concept_relation_ids (List[str], optional): List of concept relation IDs of concept relations.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> app.delete_concept_relations(concept_id="concept_id", concept_relation_ids=["concept_relation_id_1", "concept_relation_id_2", ..])
        """
        if not concept_relation_ids:
            concept_relation_ids = [
                concept_relation.id
                for concept_relation in list(self.search_concept_relations(concept_id=concept_id))
            ]
        request = service_pb2.DeleteConceptRelationsRequest(
            user_app_id=self.user_app_id, concept_id=concept_id, ids=concept_relation_ids
        )
        response = self._grpc_request(self.STUB.DeleteConceptRelations, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nConcept Relations Deleted\n%s", response.status)

    def search(self, **kwargs) -> Model:
        """Returns a Search object for the user and app ID.

        Args:
            see the Search class in clarifai.client.search for kwargs.

        Returns:
            Search: A Search object for the user and app ID.

        Example:
            >>> from clarifai.client.app import App
            >>> app = App(app_id="app_id", user_id="user_id")
            >>> search_client = app.search(top_k=12, metric="euclidean")
        """
        kwargs.get("user_id", self.user_app_id.user_id)
        kwargs.get("app_id", self.user_app_id.app_id)
        return Search.from_auth_helper(auth=self.auth_helper, **kwargs)

    def __getattr__(self, name):
        return getattr(self.app_info, name)

    def __str__(self):
        init_params = [param for param in self.kwargs.keys()]
        attribute_strings = [
            f"{param}={getattr(self.app_info, param)}"
            for param in init_params
            if hasattr(self.app_info, param)
        ]
        return f"Clarifai App Details: \n{', '.join(attribute_strings)}\n"



import os
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Type

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct

from clarifai.client.input import Inputs
from clarifai.datasets.upload.base import ClarifaiDataLoader, ClarifaiDataset
from clarifai.utils.misc import get_uuid


class VisualClassificationDataset(ClarifaiDataset):
    def __init__(
        self, data_generator: Type[ClarifaiDataLoader], dataset_id: str, max_workers: int = 4
    ) -> None:
        super().__init__(data_generator, dataset_id, max_workers)

    def _extract_protos(
        self, batch_input_ids: List[str]
    ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
        """Create input image and annotation protos for batch of input ids.
        Args:
          batch_input_ids: List of input IDs to retrieve the protos for.
        Returns:
          input_protos: List of input protos.
          annotation_protos: List of annotation protos.
        """
        input_protos, annotation_protos = [], []

        def process_data_item(id):
            data_item = self.data_generator[id]
            metadata = Struct()
            image_path = data_item.image_path
            labels = (
                data_item.labels if isinstance(data_item.labels, list) else [data_item.labels]
            )  # clarifai concept
            label_ids = data_item.label_ids
            input_id = (
                f"{self.dataset_id}-{get_uuid(8)}"
                if data_item.id is None
                else f"{self.dataset_id}-{str(data_item.id)}"
            )
            geo_info = data_item.geo_info
            if data_item.metadata is not None:
                metadata.update(data_item.metadata)
            elif image_path is not None:
                metadata.update({"filename": os.path.basename(image_path)})
            else:
                metadata = None

            self.all_input_ids[id] = input_id
            if data_item.image_bytes is not None:
                input_protos.append(
                    Inputs.get_input_from_bytes(
                        input_id=input_id,
                        image_bytes=data_item.image_bytes,
                        dataset_id=self.dataset_id,
                        labels=labels,
                        label_ids=label_ids,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )
            else:
                input_protos.append(
                    Inputs.get_input_from_file(
                        input_id=input_id,
                        image_file=image_path,
                        dataset_id=self.dataset_id,
                        labels=labels,
                        label_ids=label_ids,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
            for job in futures:
                job.result()

        return input_protos, annotation_protos


class VisualDetectionDataset(ClarifaiDataset):
    """Visual detection dataset proto class."""

    def __init__(
        self, data_generator: Type[ClarifaiDataLoader], dataset_id: str, max_workers: int = 4
    ) -> None:
        super().__init__(data_generator, dataset_id, max_workers)

    def _extract_protos(
        self, batch_input_ids: List[int]
    ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
        """Create input image protos for each data generator item.
        Args:
          batch_input_ids: List of input IDs to retrieve the protos for.
        Returns:
          input_protos: List of input protos.
          annotation_protos: List of annotation protos.
        """
        input_protos, annotation_protos = [], []

        def process_data_item(id):
            data_item = self.data_generator[id]
            metadata = Struct()
            image = data_item.image_path
            labels = data_item.labels  # list:[l1,...,ln]
            if data_item.label_ids is not None:
                assert len(labels) == len(data_item.label_ids), (
                    "Length of labels and label_ids must be equal"
                )
                label_ids = data_item.label_ids
            else:
                label_ids = None
            bboxes = data_item.bboxes  # [[xmin,ymin,xmax,ymax],...,[xmin,ymin,xmax,ymax]]
            input_id = (
                f"{self.dataset_id}-{get_uuid(8)}"
                if data_item.id is None
                else f"{self.dataset_id}-{str(data_item.id)}"
            )
            if data_item.metadata is not None:
                metadata.update(data_item.metadata)
            else:
                metadata.update({"filename": os.path.basename(image)})
            geo_info = data_item.geo_info

            self.all_input_ids[id] = input_id
            if data_item.image_bytes is not None:
                input_protos.append(
                    Inputs.get_input_from_bytes(
                        input_id=input_id,
                        image_bytes=data_item.image_bytes,
                        dataset_id=self.dataset_id,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )
            else:
                input_protos.append(
                    Inputs.get_input_from_file(
                        input_id=input_id,
                        image_file=image,
                        dataset_id=self.dataset_id,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )
            # iter over bboxes and labels
            # one id could have more than one bbox and label
            for i in range(len(bboxes)):
                annotation_protos.append(
                    Inputs.get_bbox_proto(
                        input_id=input_id,
                        label=labels[i],
                        bbox=bboxes[i],
                        label_id=label_ids[i] if label_ids else None,
                    )
                )

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
            for job in futures:
                job.result()

        return input_protos, annotation_protos


class VisualSegmentationDataset(ClarifaiDataset):
    """Visual segmentation dataset proto class."""

    def __init__(
        self, data_generator: Type[ClarifaiDataLoader], dataset_id: str, max_workers: int = 4
    ) -> None:
        super().__init__(data_generator, dataset_id, max_workers)

    def _extract_protos(
        self, batch_input_ids: List[str]
    ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
        """Create input image and annotation protos for batch of input ids.
        Args:
          batch_input_ids: List of input IDs to retrieve the protos for.
        Returns:
          input_protos: List of input protos.
          annotation_protos: List of annotation protos.
        """
        input_protos, annotation_protos = [], []

        def process_data_item(id):
            data_item = self.data_generator[id]
            metadata = Struct()
            image = data_item.image_path
            labels = data_item.labels
            if data_item.label_ids is not None:
                assert len(labels) == len(data_item.label_ids), (
                    "Length of labels and label_ids must be equal"
                )
                label_ids = data_item.label_ids
            else:
                label_ids = None
            _polygons = data_item.polygons  # list of polygons: [[[x,y],...,[x,y]],...]
            input_id = (
                f"{self.dataset_id}-{get_uuid(8)}"
                if data_item.id is None
                else f"{self.dataset_id}-{str(data_item.id)}"
            )
            if data_item.metadata is not None:
                metadata.update(data_item.metadata)
            else:
                metadata.update({"filename": os.path.basename(image)})
            geo_info = data_item.geo_info

            self.all_input_ids[id] = input_id
            if data_item.image_bytes is not None:
                input_protos.append(
                    Inputs.get_input_from_bytes(
                        input_id=input_id,
                        image_bytes=data_item.image_bytes,
                        dataset_id=self.dataset_id,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )
            else:
                input_protos.append(
                    Inputs.get_input_from_file(
                        input_id=input_id,
                        image_file=image,
                        dataset_id=self.dataset_id,
                        geo_info=geo_info,
                        metadata=metadata,
                    )
                )

            ## Iterate over each masked image and create a proto for upload to clarifai
            ## The length of masks/polygons-list and labels must be equal
            for i, _polygon in enumerate(_polygons):
                try:
                    annotation_protos.append(
                        Inputs.get_mask_proto(
                            input_id=input_id,
                            label=labels[i],
                            polygons=_polygon,
                            label_id=label_ids[i] if label_ids else None,
                        )
                    )
                except IndexError:
                    continue

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
            for job in futures:
                job.result()

        return input_protos, annotation_protos



from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import DEFAULT_BASE


class PipelineStep(Lister, BaseClient):
    """PipelineStep is a class that provides access to Clarifai API endpoints related to PipelineStep information."""

    def __init__(
        self,
        url: str = None,
        pipeline_step_id: str = None,
        pipeline_step_version_id: str = None,
        user_id: str = None,
        app_id: str = None,
        pipeline_id: str = None,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
        **kwargs,
    ):
        """Initializes a PipelineStep object.

        Args:
            url (str): The URL to initialize the pipeline step object.
            pipeline_step_id (str): The PipelineStep ID for the PipelineStep to interact with.
            pipeline_step_version_id (str): The PipelineStep version ID for the PipelineStep to interact with.
            user_id (str): The User ID for the PipelineStep to interact with.
            app_id (str): The App ID for the PipelineStep to interact with.
            pipeline_id (str): The Pipeline ID for the PipelineStep to interact with.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication.
            token (str): A session token for authentication.
            root_certificates_path (str): Path to the SSL root certificates file.
            **kwargs: Additional keyword arguments to be passed to the BaseClient.
        """
        if url:
            user_id, app_id, _, pipeline_step_id, pipeline_step_version_id = (
                ClarifaiUrlHelper.split_clarifai_url(url)
            )

        # Store all kwargs as attributes for API data
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.kwargs = {
            "pipeline_step_id": pipeline_step_id,
            "pipeline_step_version_id": pipeline_step_version_id,
            "user_id": user_id,
            "app_id": app_id,
            "pipeline_id": pipeline_id,
            **kwargs,
        }

        BaseClient.__init__(
            self,
            user_id=user_id,
            app_id=app_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

        self.pipeline_step_id = pipeline_step_id
        self.pipeline_step_version_id = pipeline_step_version_id
        self.pipeline_id = pipeline_id
        self.user_id = user_id
        self.app_id = app_id



import base64
import json
import math
import operator
from io import BytesIO
from typing import Dict, List

import requests
from clarifai_grpc.grpc.api import resources_pb2
from clarifai_grpc.grpc.api.resources_pb2 import ModelTypeEnumOption, ModelTypeRangeInfo
from clarifai_grpc.grpc.api.resources_pb2 import ModelTypeField as ParamProto
from PIL import Image as PILImage

from clarifai.runners.utils.data_types import Audio, Image, MessageData, Video


def image_to_bytes(img: PILImage.Image, format="JPEG") -> bytes:
    buffered = BytesIO()
    img.save(buffered, format=format)
    img_str = buffered.getvalue()
    return img_str


def bytes_to_image(bytes_img) -> PILImage.Image:
    img = PILImage.open(BytesIO(bytes_img))
    return img


def process_image(image: Image) -> Dict:
    """Convert Clarifai Image object to OpenAI image format."""

    if image.bytes:
        b64_img = image.to_base64_str()
        return {'type': 'image_url', 'image_url': {'url': f"data:image/jpeg;base64,{b64_img}"}}
    elif image.url:
        return {'type': 'image_url', 'image_url': {'url': image.url}}
    else:
        raise ValueError("Image must contain either bytes or URL")


def process_audio(audio: Audio) -> Dict:
    """Convert Clarifai Audio object to OpenAI audio format."""

    if audio.bytes:
        audio = audio.to_base64_str()
        audio = {
            "type": "input_audio",
            "input_audio": {"data": audio, "format": "wav"},
        }
    elif audio.url:
        response = requests.get(audio.url)
        if response.status_code != 200:
            raise ValueError(f"Failed to fetch audio from URL: {audio.url}")
        audio_base64_str = base64.b64encode(response.content).decode('utf-8')
        audio = {
            "type": "input_audio",
            "input_audio": {"data": audio_base64_str, "format": "wav"},
        }
    else:
        raise ValueError("Audio must contain either bytes or URL")

    return audio


def process_video(video: Video) -> Dict:
    """Convert Clarifai Video object to OpenAI video format."""

    if video.bytes:
        video = "data:video/mp4;base64," + video.to_base64_str()
        video = {
            "type": "video_url",
            "video_url": {"url": video},
        }
    elif video.url:
        response = requests.get(video.url)
        if response.status_code != 200:
            raise ValueError(f"Failed to fetch video from URL: {video.url}")
        video_base64_str = base64.b64encode(response.content).decode('utf-8')
        video = {
            "type": "video_url",
            "video_url": {"url": video_base64_str},
        }
    else:
        raise ValueError("Video must contain either bytes or URL")

    return video


class Param(MessageData):
    """A field that can be used to store input data."""

    def __init__(
        self,
        default,
        description=None,
        min_value=None,
        max_value=None,
        choices=None,
        is_param=True,
    ):
        self.default = default
        self.description = description
        self.min_value = min_value
        self.max_value = max_value
        self.choices = choices
        self.is_param = is_param
        self._patch_encoder()

    def __repr__(self) -> str:
        attrs = []
        if self.default is not None:
            attrs.append(f"default={self.default!r}")
        if self.description is not None:
            attrs.append(f"description={self.description!r}")
        if self.min_value is not None:
            attrs.append(f"min_value={self.min_value!r}")
        if self.max_value is not None:
            attrs.append(f"max_value={self.max_value!r}")
        if self.choices is not None:
            attrs.append(f"choices={self.choices!r}")
        attrs.append(f"is_param={self.is_param!r}")
        return f"Param({', '.join(attrs)})"

    # All *explicit* conversions
    def __int__(self):
        return int(self.default)

    def __float__(self):
        return float(self.default)

    def __str__(self):
        return str(self.default)

    def __bool__(self):
        return bool(self.default)

    def __index__(self):
        return int(self.default)  # for slicing

    # sequence / mapping protocol delegation
    def __len__(self):
        return len(self.default)

    def __iter__(self):
        return iter(self.default)

    def __reversed__(self):
        return reversed(self.default)

    def __contains__(self, item):
        return item in self.default

    def __getitem__(self, key):
        return self.default[key]

    def __setitem__(self, k, v):
        self.default[k] = v

    def __delitem__(self, k):
        del self.default[k]

    def __hash__(self):
        return hash(self.default)

    def __call__(self, *args, **kwargs):
        return self.default(*args, **kwargs)

    # Comparison operators
    def __eq__(self, other):
        return self.default == other

    def __lt__(self, other):
        return self.default < other

    def __le__(self, other):
        return self.default <= other

    def __gt__(self, other):
        return self.default > other

    def __ge__(self, other):
        return self.default >= other

    def __getattribute__(self, name):
        """Intercept attribute access to mimic default value behavior"""
        try:
            # First try to get Param attributes normally
            return object.__getattribute__(self, name)
        except AttributeError:
            # Fall back to the default value's attributes
            default = object.__getattribute__(self, 'default')
            return getattr(default, name)

    # Arithmetic operators – # arithmetic & bitwise operators – auto-generated
    _arith_ops = {
        "__add__": operator.add,
        "__sub__": operator.sub,
        "__mul__": operator.mul,
        "__truediv__": operator.truediv,
        "__floordiv__": operator.floordiv,
        "__mod__": operator.mod,
        "__pow__": operator.pow,
        "__and__": operator.and_,
        "__or__": operator.or_,
        "__xor__": operator.xor,
        "__lshift__": operator.lshift,
        "__rshift__": operator.rshift,
    }

    for _name, _op in _arith_ops.items():

        def _make(op):
            def _f(self, other, *, _op=op):  # default arg binds op
                return _op(self.default, other)

            return _f

        locals()[_name] = _make(_op)
        locals()["__r" + _name[2:]] = _make(lambda x, y, _op=_op: _op(y, x))
    del _name, _op, _make

    # In-place operators
    _inplace_ops = {
        "__iadd__": operator.iadd,
        "__isub__": operator.isub,
        "__imul__": operator.imul,
        "__itruediv__": operator.itruediv,
        "__ifloordiv__": operator.ifloordiv,
        "__imod__": operator.imod,
        "__ipow__": operator.ipow,
        "__ilshift__": operator.ilshift,
        "__irshift__": operator.irshift,
        "__iand__": operator.iand,
        "__ixor__": operator.ixor,
        "__ior__": operator.ior,
    }

    for _name, _op in _inplace_ops.items():

        def _make_inplace(op):
            def _f(self, other, *, _op=op):
                self.default = _op(self.default, other)
                return self

            return _f

        locals()[_name] = _make_inplace(_op)
    del _name, _op, _make_inplace

    # Formatting and other conversions
    def __format__(self, format_spec):
        return format(self.default, format_spec)

    def __bytes__(self):
        return bytes(self.default)

    def __complex__(self):
        return complex(self.default)

    def __round__(self, ndigits=None):
        return round(self.default, ndigits)

    def __trunc__(self):
        return math.trunc(self.default)

    def __floor__(self):
        return math.floor(self.default)

    def __ceil__(self):
        return math.ceil(self.default)

    # Attribute access delegation – anything we did *not* define above
    # will automatically be looked up on the wrapped default value.
    # Attribute access delegation
    def __getattr__(self, item):
        return getattr(self.default, item)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.default

    def __json__(self):
        return self.default if not hasattr(self.default, '__json__') else self.default.__json__()

    @classmethod
    def _patch_encoder(cls):
        # only patch once
        if getattr(json.JSONEncoder, "_user_patched", False):
            return
        original = json.JSONEncoder.default

        def default(self, obj):
            if isinstance(obj, Param):
                return obj.__json__()
            return original(self, obj)

        json.JSONEncoder.default = default
        json.JSONEncoder._user_patched = True

    def to_proto(self, proto=None) -> ParamProto:
        if proto is None:
            proto = ParamProto()
        if self.description is not None:
            proto.description = self.description

        if self.choices is not None:
            for choice in self.choices:
                option = ModelTypeEnumOption(id=str(choice))
                proto.model_type_enum_options.append(option)

        proto.required = False

        if self.min_value is not None or self.max_value is not None:
            range_info = ModelTypeRangeInfo()
            if self.min_value is not None:
                range_info.min = float(self.min_value)
            if self.max_value is not None:
                range_info.max = float(self.max_value)
            proto.model_type_range_info.CopyFrom(range_info)
        proto.is_param = self.is_param

        proto = self.set_default(proto, self.default)
        return proto

    @classmethod
    def from_proto(cls, proto):
        default = None
        if proto.HasField('default'):
            pb_value = proto.default
            if pb_value.HasField('string_value'):
                default = pb_value.string_value
                try:
                    import json

                    default = json.loads(default)
                except json.JSONDecodeError:
                    pass
            elif pb_value.HasField('number_value'):
                default = pb_value.number_value
                if default.is_integer():
                    default = int(default)
                else:
                    default = float(default)
            elif pb_value.HasField('bool_value'):
                default = pb_value.bool_value

        choices = (
            [option.id for option in proto.model_type_enum_options]
            if proto.model_type_enum_options
            else None
        )

        min_value = None
        max_value = None
        if proto.HasField('model_type_range_info'):
            min_value = proto.model_type_range_info.min
            max_value = proto.model_type_range_info.max
            if min_value.is_integer():
                min_value = int(min_value)
            if max_value.is_integer():
                max_value = int(max_value)

        return cls(
            default=default,
            description=proto.description if proto.description else None,
            min_value=min_value,
            max_value=max_value,
            choices=choices,
            is_param=proto.is_param,
        )

    @classmethod
    def set_default(cls, proto=None, default=None):
        try:
            import json

            if proto is None:
                proto = ParamProto()

            if isinstance(default, str):
                proto.default = default
            else:
                proto.default = json.dumps(default)

            return proto

        except Exception:
            if default is not None:
                proto.default = str(default)
            return proto
        except Exception as e:
            raise ValueError(
                f"Error setting default value of type, {type(default)} and value: {default}: {e}"
            )

    @classmethod
    def get_default(cls, proto):
        default_str = proto.default
        default = None
        import json

        try:
            # Attempt to parse as JSON first (for complex types)
            return json.loads(default_str)
        except json.JSONDecodeError:
            pass
        # Check for boolean values stored as "True" or "False"
        if proto.type == resources_pb2.ModelTypeField.DataType.BOOL:
            try:
                default = bool(default_str)
            except ValueError:
                pass
        # Try to parse as integer
        elif proto.type == resources_pb2.ModelTypeField.DataType.INT:
            try:
                default = int(default_str)
            except ValueError:
                pass

        # Try to parse as float
        elif proto.type == resources_pb2.ModelTypeField.DataType.FLOAT:
            try:
                default = float(default_str)
            except ValueError:
                pass
        elif proto.type == resources_pb2.ModelTypeField.DataType.STR:
            default = default_str

        if default is None:
            # If all parsing fails, return the string value
            default = default_str
        return default


class DataConverter:
    """A class that can be used to convert data to and from a specific format."""

    @classmethod
    def convert_input_data_to_new_format(
        cls, data: resources_pb2.Data, input_fields: List[resources_pb2.ModelTypeField]
    ) -> resources_pb2.Data:
        """Convert input data to new format."""
        new_data = resources_pb2.Data()
        for field in input_fields:
            part_data = cls._convert_field(data, field)
            if cls._is_data_set(part_data):
                # if the field is set, add it to the new data part
                part = new_data.parts.add()
                part.id = field.name
                part.data.CopyFrom(part_data)
            elif field.required:
                raise ValueError(f"Field {field.name} is required but not set")
        return new_data

    @classmethod
    def _convert_field(
        cls, old_data: resources_pb2.Data, field: resources_pb2.ModelTypeField
    ) -> resources_pb2.Data:
        data_type = field.type
        new_data = resources_pb2.Data()
        if data_type == resources_pb2.ModelTypeField.DataType.STR:
            if old_data.HasField('text'):
                new_data.string_value = old_data.text.raw
                old_data.ClearField('text')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.IMAGE:
            if old_data.HasField('image'):
                new_data.image.CopyFrom(old_data.image)
                # Clear the old field to avoid duplication
                old_data.ClearField('image')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.VIDEO:
            if old_data.HasField('video'):
                new_data.video.CopyFrom(old_data.video)
                old_data.ClearField('video')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.BOOL:
            if old_data.bool_value is not False:
                new_data.bool_value = old_data.bool_value
                old_data.bool_value = False
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.INT:
            if old_data.int_value != 0:
                new_data.int_value = old_data.int_value
                old_data.int_value = 0
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.FLOAT:
            if old_data.float_value != 0.0:
                new_data.float_value = old_data.float_value
                old_data.float_value = 0.0
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.BYTES:
            if old_data.bytes_value != b"":
                new_data.bytes_value = old_data.bytes_value
                old_data.bytes_value = b""
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.NDARRAY:
            if old_data.HasField('ndarray'):
                new_data.ndarray.CopyFrom(old_data.ndarray)
                old_data.ClearField('ndarray')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.TEXT:
            if old_data.HasField('text'):
                new_data.text.CopyFrom(old_data.text)
                old_data.ClearField('text')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.AUDIO:
            if old_data.HasField('audio'):
                new_data.audio.CopyFrom(old_data.audio)
                old_data.ClearField('audio')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.CONCEPT:
            if old_data.concepts:
                new_data.concepts.extend(old_data.concepts)
                old_data.ClearField('concepts')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.REGION:
            if old_data.regions:
                new_data.regions.extend(old_data.regions)
                old_data.ClearField('regions')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.FRAME:
            if old_data.frames:
                new_data.frames.extend(old_data.frames)
                old_data.ClearField('frames')
            return new_data
        elif data_type == resources_pb2.ModelTypeField.DataType.LIST:
            if not field.type_args:
                raise ValueError("LIST type requires type_args")
            element_field = field.type_args[0]
            if element_field in (
                resources_pb2.ModelTypeField.DataType.CONCEPT,
                resources_pb2.ModelTypeField.DataType.REGION,
                resources_pb2.ModelTypeField.DataType.FRAME,
            ):
                # convert to new format
                new_data = cls._convert_field(old_data, element_field)
            return new_data
        else:
            return new_data
            # raise ValueError(f"Unsupported data type: {data_type}")

    @classmethod
    def is_old_format(cls, data: resources_pb2.Data) -> bool:
        """Check if the Data proto is in the old format (without parts)."""
        if len(data.parts) > 0:
            return False  # New format uses parts

        # Check if any singular field is set
        singular_fields = [
            'image',
            'video',
            'metadata',
            'geo',
            'text',
            'audio',
            'ndarray',
            'int_value',
            'float_value',
            'bytes_value',
            'bool_value',
            'string_value',
        ]
        for field in singular_fields:
            if data.HasField(field):
                return True

        # Check if any repeated field has elements
        repeated_fields = [
            'concepts',
            'colors',
            'clusters',
            'embeddings',
            'regions',
            'frames',
            'tracks',
            'time_segments',
            'hits',
            'heatmaps',
        ]
        for field in repeated_fields:
            if getattr(data, field):
                return True

        return False

    @classmethod
    def convert_output_data_to_old_format(cls, data: resources_pb2.Data) -> resources_pb2.Data:
        """Convert output data to old format."""
        old_data = resources_pb2.Data()
        part_data = data.parts[0].data
        # Handle text.raw specially (common case for text outputs)
        old_data = part_data
        if old_data.string_value:
            old_data.text.raw = old_data.string_value

        return old_data

    @classmethod
    def _is_data_set(cls, data_msg):
        # Singular message fields
        singular_fields = ["image", "video", "metadata", "geo", "text", "audio", "ndarray"]
        for field in singular_fields:
            if data_msg.HasField(field):
                return True

        # Repeated fields
        repeated_fields = [
            "concepts",
            "colors",
            "clusters",
            "embeddings",
            "regions",
            "frames",
            "tracks",
            "time_segments",
            "hits",
            "heatmaps",
            "parts",
        ]
        for field in repeated_fields:
            if getattr(data_msg, field):  # checks if the list is not empty
                return True

        # Scalar fields (proto3 default: 0 for numbers, empty for strings/bytes, False for bool)
        if (
            data_msg.int_value != 0
            or data_msg.float_value != 0.0
            or data_msg.bytes_value != b""
            or data_msg.bool_value is True
            or data_msg.string_value != ""
        ):
            return True

        return False



import asyncio
import itertools
import time
from concurrent.futures import ThreadPoolExecutor

import grpc
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.client.auth.register import RpcCallable, V2Stub
from clarifai.utils.logging import logger

throttle_status_codes = {
    status_code_pb2.CONN_THROTTLED,
    status_code_pb2.CONN_EXCEED_HOURLY_LIMIT,
}

retry_codes_grpc = {
    grpc.StatusCode.UNAVAILABLE,
}

_threadpool = ThreadPoolExecutor(100)


def validate_response(response, attempt, max_attempts):
    # Helper function to handle simple response validation
    def handle_simple_response(response):
        if hasattr(response, 'status') and hasattr(response.status, 'code'):
            if (response.status.code in throttle_status_codes) and attempt < max_attempts:
                logger.debug('Retrying with status %s' % str(response.status))
                return None  # Indicates a retry is needed
            else:
                return response

        # Check if response is an async iterator
        if hasattr(response, '__aiter__'):
            return response  # Return async iterator directly for handling in _async_call__

    # Check if the response is an instance of a gRPC streaming call
    if isinstance(response, grpc._channel._MultiThreadedRendezvous):
        try:
            # Check just the first response in the stream for validation
            first_res = next(response)
            validated_response = handle_simple_response(first_res)
            if validated_response is not None:
                # Have to return that first response and the rest of the stream.
                return itertools.chain([validated_response], response)
            return None  # Indicates a retry is needed
        except grpc.RpcError as e:
            logger.error('Error processing streaming response: %s' % str(e))
            return None  # Indicates an error
    else:
        # Handle simple response validation
        return handle_simple_response(response)


def create_stub(
    auth_helper: ClarifaiAuthHelper = None, max_retry_attempts: int = 10, is_async: bool = False
) -> V2Stub:
    """
    Create client stub that handles authorization and basic retries for
    unavailable or throttled connections.

    Args:
      auth_helper:  ClarifaiAuthHelper to use for auth metadata (default: from env)
      max_retry_attempts:  max attempts to retry rpcs with retryable failures
    """
    stub = AuthorizedStub(auth_helper, is_async=is_async)
    if max_retry_attempts > 0:
        return RetryStub(stub, max_retry_attempts, is_async=is_async)
    return stub


class AuthorizedStub(V2Stub):
    """V2Stub proxy that inserts metadata authorization in rpc calls."""

    def __init__(self, auth_helper: ClarifaiAuthHelper = None, is_async: bool = False):
        if auth_helper is None:
            auth_helper = ClarifaiAuthHelper.from_env()
        self.is_async = is_async
        self.stub = auth_helper.get_async_stub() if is_async else auth_helper.get_stub()
        self.metadata = auth_helper.metadata

    def __getattr__(self, name):
        value = getattr(self.stub, name)
        if isinstance(value, RpcCallable):
            value = _AuthorizedRpcCallable(value, self.metadata, self.is_async)
        return value


class _AuthorizedRpcCallable(RpcCallable):
    """Adds metadata(authorization header) to rpc calls"""

    def __init__(self, func, metadata, is_async):
        self.f = func
        self.metadata = metadata
        self.is_async = is_async or asyncio.iscoroutinefunction(func)

    def __repr__(self):
        return repr(self.f)

    def __call__(self, *args, **kwargs):
        metadata = kwargs.pop('metadata', self.metadata)

        return self.f(
            *args,
            **kwargs,
            metadata=metadata,
        )

    def future(self, *args, **kwargs):
        metadata = kwargs.pop('metadata', self.metadata)
        return self.f.future(*args, **kwargs, metadata=metadata)

    def __getattr__(self, name):
        return getattr(self.f, name)


class RetryStub(V2Stub):
    """
    V2Stub proxy that retries requests (currently on unavailable server or throttle codes)
    """

    def __init__(self, stub, max_attempts=10, backoff_time=5, is_async=False):
        self.stub = stub
        self.max_attempts = max_attempts
        self.backoff_time = backoff_time
        self.is_async = is_async

    def __getattr__(self, name):
        value = getattr(self.stub, name)
        if isinstance(value, RpcCallable):
            value = _RetryRpcCallable(value, self.max_attempts, self.backoff_time, self.is_async)
        return value


class _RetryRpcCallable(RpcCallable):
    """Retries rpc calls on unavailable server or throttle codes"""

    def __init__(self, func, max_attempts, backoff_time, is_async=False):
        self.f = func
        self.max_attempts = max_attempts
        self.backoff_time = backoff_time
        self.is_async = is_async or asyncio.iscoroutinefunction(func)

    def __repr__(self):
        return repr(self.f)

    async def _async_call__(self, *args, **kwargs):
        """Handle async RPC calls with retries and validation"""
        for attempt in range(1, self.max_attempts + 1):
            if attempt != 1:
                await asyncio.sleep(self.backoff_time)

            try:
                response = self.f(*args, **kwargs)

                # Handle streaming response
                if hasattr(response, '__aiter__'):
                    return await self._handle_streaming_response(response, attempt)

                # Handle regular async response
                validated_response = await self._handle_regular_response(response, attempt)
                if validated_response is not None:
                    return validated_response

            except grpc.RpcError as e:
                if not self._should_retry(e, attempt):
                    raise
                logger.debug(
                    f'Retrying after error {e.code()} (attempt {attempt}/{self.max_attempts})'
                )

        raise Exception(f'Max attempts reached ({self.max_attempts}) without success')

    async def _handle_streaming_response(self, response, attempt):
        """Handle streaming response validation and processing"""

        async def validated_stream():
            try:
                async for item in response:
                    if not self._is_valid_response(item):
                        if attempt < self.max_attempts:
                            yield None  # Signal for retry
                        raise Exception(
                            f'Validation failed on streaming response (attempt {attempt})'
                        )
                    yield item
            except grpc.RpcError as e:
                if not self._should_retry(e, attempt):
                    raise
                yield None  # Signal for retry

        return validated_stream()

    async def _handle_regular_response(self, response, attempt):
        """Handle regular async response validation and processing"""
        try:
            result = await response
            if not self._is_valid_response(result):
                if attempt < self.max_attempts:
                    return None  # Signal for retry
                raise Exception(f'Validation failed on response (attempt {attempt})')
            return result
        except grpc.RpcError as e:
            if not self._should_retry(e, attempt):
                raise
            return None  # Signal for retry

    def _is_valid_response(self, response):
        """Check if response status is valid"""
        return not (
            hasattr(response, 'status')
            and hasattr(response.status, 'code')
            and response.status.code in throttle_status_codes
        )

    def _should_retry(self, error, attempt):
        """Determine if we should retry based on error and attempt count"""
        return (
            isinstance(error, grpc.RpcError)
            and error.code() in retry_codes_grpc
            and attempt < self.max_attempts
        )

    def _sync_call__(self, *args, **kwargs):
        attempt = 0
        while attempt < self.max_attempts:
            attempt += 1
            if attempt != 1:
                time.sleep(self.backoff_time)  # TODO better backoff between attempts
            try:
                response = self.f(*args, **kwargs)
                v = validate_response(response, attempt, self.max_attempts)
                if v is not None:
                    return v
            except grpc.RpcError as e:
                if (e.code() in retry_codes_grpc) and attempt < self.max_attempts:
                    logger.debug('Retrying with status %s' % e.code())
                else:
                    raise

    def __call__(self, *args, **kwargs):
        if self.is_async:
            return self._async_call__(*args, **kwargs)
        return self._sync_call__(*args, **kwargs)

    async def __call_async__(self, *args, **kwargs):
        """Explicit async call method"""
        return await self._async_call(*args, **kwargs)

    def future(self, *args, **kwargs):
        # TODO use single result event loop thread with asyncio
        if self.is_async:
            return asyncio.create_task(self._async_call(*args, **kwargs))
        return _threadpool.submit(self, *args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.f, name)



import glob
import json
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from typing import DefaultDict, Dict, List

from PIL import Image
from tqdm import tqdm

from clarifai.datasets.upload.base import ClarifaiDataLoader

from ..features import VisualDetectionFeatures


class xviewDetectionDataLoader(ClarifaiDataLoader):
    """xview Image Detection Dataset"""

    xview_concepts = [
        'Fixed-wing Aircraft',
        'Small Aircraft',
        'Cargo Plane',
        'Helicopter',
        'Passenger Vehicle',
        'Small Car',
        'Bus',
        'Pickup Truck',
        'Utility Truck',
        'Truck',
        'Cargo Truck',
        'Truck w-Box',
        'Truck Tractor',
        'Trailer',
        'Truck w-Flatbed',
        'Truck w-Liquid',
        'Crane Truck',
        'Railway Vehicle',
        'Passenger Car',
        'Cargo Car',
        'Flat Car',
        'Tank car',
        'Locomotive',
        'Maritime Vessel',
        'Motorboat',
        'Sailboat',
        'Tugboat',
        'Barge',
        'Fishing Vessel',
        'Ferry',
        'Yacht',
        'Container Ship',
        'Oil Tanker',
        'Engineering Vehicle',
        'Tower crane',
        'Container Crane',
        'Reach Stacker',
        'Straddle Carrier',
        'Mobile Crane',
        'Dump Truck',
        'Haul Truck',
        'Scraper-Tractor',
        'Front loader-Bulldozer',
        'Excavator',
        'Cement Mixer',
        'Ground Grader',
        'Hut-Tent',
        'Shed',
        'Building',
        'Aircraft Hangar',
        'Damaged Building',
        'Facility',
        'Construction Site',
        'Vehicle Lot',
        'Helipad',
        'Storage Tank',
        'Shipping container lot',
        'Shipping Container',
        'Pylon',
        'Tower',
    ]

    def __init__(self, data_dir) -> None:
        """Initialize and Compress xview dataset.
        Args:
        data_dir: the local dataset directory.
        """

        self.data_dir = data_dir
        self.img_dir = os.path.join(self.data_dir, "train_images")
        self.img_comp_dir = os.path.join(self.data_dir, "train_images_comp")
        self.label_file = os.path.join(self.data_dir, "xview_train.geojson")

        self.preprocess()
        self.all_data = self.xview_data_parser()

        self.load_data()

    @property
    def task(self):
        return "visual_detection"

    def compress_tiff(self, img_path: str) -> None:
        """Compress tiff image"""
        img_comp_path = os.path.join(self.img_comp_dir, os.path.basename(img_path))
        img_arr = Image.open(img_path)
        img_arr.save(img_comp_path, 'TIFF', compression='tiff_deflate')

    def preprocess(self):
        """Compress the tiff images to comply with clarifai grpc image encoding limit(<20MB) Uses ADOBE_DEFLATE compression algorithm"""
        all_img_ids = glob.glob(os.path.join(self.img_dir, "*.tif"))

        if not os.path.exists(self.img_comp_dir):
            os.mkdir(self.img_comp_dir)

        num_workers = cpu_count()
        futures = []

        with ThreadPoolExecutor(max_workers=num_workers) as executor:
            with tqdm(total=len(all_img_ids), desc="compressing ...") as progress:
                for img_path in all_img_ids:
                    future = executor.submit(self.compress_tiff, img_path)
                    future.add_done_callback(lambda _: progress.update())
                    futures.append(future)

                results = []
                for future in futures:
                    result = future.result()
                    results.append(result)

    def xview_classes2indices(self, classes: List) -> List:
        """remap xview classes 11-94 to 0-59"""
        indices = [
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            0,
            1,
            2,
            -1,
            3,
            -1,
            4,
            5,
            6,
            7,
            8,
            -1,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            -1,
            -1,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            -1,
            23,
            24,
            25,
            -1,
            26,
            27,
            -1,
            28,
            -1,
            29,
            30,
            31,
            32,
            33,
            34,
            35,
            36,
            37,
            -1,
            38,
            39,
            40,
            41,
            42,
            43,
            44,
            45,
            -1,
            -1,
            -1,
            -1,
            46,
            47,
            48,
            49,
            -1,
            50,
            51,
            -1,
            52,
            -1,
            -1,
            -1,
            53,
            54,
            -1,
            55,
            -1,
            -1,
            56,
            -1,
            57,
            -1,
            58,
            59,
        ]
        return [indices[int(c)] for c in classes]

    def xview_indices2concepts(self, indices: List) -> List:
        """remap classes to concept names"""
        return [self.xview_concepts[i] for i in indices]

    def xview_data_parser(self) -> DefaultDict[str, Dict[List, List]]:
        """Parse geojson data into nested dict of imageid w.r.t bounding boxes, concepts"""
        all_data = defaultdict(lambda: dict(bboxes=[], concepts=[]))

        with open(self.label_file) as f:
            geojson_data = json.loads(f.read())

        for feature in tqdm(
            geojson_data['features'],
            total=len(geojson_data['features']),
            desc="Parsing geojson data",
        ):
            image_id = feature['properties']['image_id'].split(".")[0]
            xview_classid = feature['properties']['type_id']
            bbox = list(map(int, feature['properties']['bounds_imcoords'].split(",")))
            concept = self.xview_indices2concepts(self.xview_classes2indices([xview_classid]))

            all_data[image_id]['bboxes'].append(bbox)
            all_data[image_id]['concepts'].append(concept[0])

        return all_data

    def load_data(self):
        """Load image paths"""
        self.image_paths = []
        all_img_ids = glob.glob(os.path.join(self.img_comp_dir, "*.tif"))
        self.image_paths = all_img_ids

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, index: int):
        """Get dataset for a given index.
        Returns:
            VisualDetectionFeature type.
        """
        _id = os.path.splitext(os.path.basename(self.image_paths[index]))[0]
        image_path = self.image_paths[index]

        image = Image.open(image_path)
        image_width, image_height = image.size
        annots = []
        class_names = []
        for bbox, concept in zip(self.all_data[_id]['bboxes'], self.all_data[_id]['concepts']):
            x_min = max(min(bbox[0] / image_width, 1.0), 0.0)  # left_col
            y_min = max(min(bbox[1] / image_height, 1.0), 0.0)  # top_row
            x_max = max(min(bbox[2] / image_width, 1.0), 0.0)  # right_col
            y_max = max(min(bbox[3] / image_height, 1.0), 0.0)  # bottom_row
            if (x_min >= x_max) or (y_min >= y_max):
                continue
            annots.append([x for x in [x_min, y_min, x_max, y_max]])
            class_names.append(concept)

        assert len(class_names) == len(annots), (
            f"Num classes must match num bbox annotations\
        for a single image. Found {len(class_names)} classes and {len(annots)} bboxes."
        )

        return VisualDetectionFeatures(image_path, class_names, annots, id=_id)
